﻿using System;
using System.Text;
using Sandbox.ModAPI;
using VRage.Game.Components;
using VRage.ModAPI;
using Sandbox.Common.ObjectBuilders;
using VRage.ObjectBuilders;
using Sandbox.Definitions;
using VRage.Game;
using Sandbox.Game.EntityComponents;
using VRage.Game.Entity;
using Sandbox.Game;
using VRageMath;
using SpaceEngineers.Game.ModAPI;
using VRage.Game.ModAPI;
using System.Collections.Generic;
using Sandbox.Game.Entities;
using Sandbox.ModAPI.Interfaces;
using System.Linq;
using Ingame = Sandbox.ModAPI.Ingame;

namespace Cheetah.AI
{
    public enum BotTypes
    {
        None,
        Invalid,
        Station,
        Fighter,
        Freighter,
        Carrier
    }

    static public class BotFabric
    {
        static public BotBase FabricateBot(IMyCubeGrid Grid, IMyRemoteControl RC)
        {
            try
            {
                BotTypes BotType = BotBase.ReadBotType(RC);

                BotBase Bot = null;
                switch (BotType)
                {
                    case BotTypes.Fighter:
                        Bot = new FighterBot(Grid);
                        break;
                    case BotTypes.Freighter:
                        Bot = new FreighterBot(Grid);
                        break;
                    case BotTypes.Station:
                        Bot = new StationBot(Grid);
                        break;
                    default:
                        if (AISessionCore.AllowThrowingErrors) throw new Exception("Invalid bot type");
                        break;
                }

                return Bot;
            }
            catch (Exception Scrap)
            {
                Grid.LogError("BotFabric.FabricateBot", Scrap);
                return null;
            }
        }
    }

    abstract public class BotBase
    {
        public IMyCubeGrid Grid { get; protected set; }
        protected readonly IMyGridTerminalSystem Term;
        public Vector3D GridPosition
        {
            get
            {
                return Grid.GetPosition();
            }
        }
        protected float GridRadius
        {
            get
            {
                return (float)Grid.WorldVolume.Radius;
            }
        }
        protected readonly TimeSpan CalmdownTime = (!AISessionCore.Debug ? Diplomacy.DefaultPeaceTime : TimeSpan.FromMinutes(3));
        protected bool initialized
        {
            get
            {
                try
                {
                    return Grid != null;
                }
                catch (Exception Scrap)
                {
                    LogError("initialized", Scrap);
                    return false;
                }
            }
        }
        virtual public bool Initialized
        {
            get
            {
                return initialized;
            }
        }
        public MyEntityUpdateEnum Update { get; protected set; }
        public IMyRemoteControl RC { get; protected set; }
        public IMyFaction OwnerFaction { get; protected set; }

        protected string DroneNameProvider
        {
            get
            {
                return $"Drone_{RC.EntityId}";
            }
        }
        public string DroneName
        {
            get
            {
                return RC.Name;
            }
            protected set
            {
                IMyEntity entity = RC as IMyEntity;
                entity.Name = value;
                MyAPIGateway.Entities.SetEntityName(entity, true);
                //DebugWrite("DroneName_Set", $"Drone EntityName set to {RC.Name}");
            }
        }
        protected bool gridoperable
        {
            get
            {
                try
                {
                    return !Grid.MarkedForClose && !Grid.Closed && Grid.InScene;
                }
                catch (Exception Scrap)
                {
                    LogError("gridoperable", Scrap);
                    return false;
                }
            }
        }
        protected bool BotOperable = false;
        protected bool Closed;
        virtual public bool Operable
        {
            get
            {
                try
                {
                    return !Closed && Initialized && gridoperable && RC.IsFunctional && BotOperable;
                }
                catch (Exception Scrap)
                {
                    LogError("Operable", Scrap);
                    return false;
                }
            }
        }
        public List<IMyRadioAntenna> Antennae { get; protected set; }
        public delegate void OnDamageTaken(IMySlimBlock DamagedBlock, MyDamageInformation Damage);
        protected event OnDamageTaken OnDamaged;
        public delegate void hOnBlockPlaced(IMySlimBlock Block);
        protected event hOnBlockPlaced OnBlockPlaced;
        protected event Action Alert;

        public BotBase(IMyCubeGrid Grid)
        {
            if (Grid == null) return;
            this.Grid = Grid;
            Term = Grid.GetTerminalSystem();
            Antennae = new List<IMyRadioAntenna>();
        }

        static public BotTypes ReadBotType(IMyRemoteControl RC)
        {
            try
            {
                string _CustomData = RC.CustomData.Trim().Replace("\r\n", "\n");
                List<string> CustomData = new List<string>(_CustomData.Split('\n'));

                if (_CustomData.IsNullEmptyOrWhiteSpace()) return BotTypes.None;
                if (CustomData.Count < 2)
                {
                    if (AISessionCore.AllowThrowingErrors) throw new Exception("CustomData is invalid", new Exception("CustomData consists of less than two lines"));
                    else return BotTypes.Invalid;
                }
                if (CustomData[0].Trim() != "[EEM_AI]")
                {
                    if (AISessionCore.AllowThrowingErrors) throw new Exception("CustomData is invalid", new Exception($"AI tag invalid: '{CustomData[0]}'"));
                    return BotTypes.Invalid;
                }

                string[] bottype = CustomData[1].Split(':');
                if (bottype[0].Trim() != "Type")
                {
                    if (AISessionCore.AllowThrowingErrors) throw new Exception("CustomData is invalid", new Exception($"Type tag invalid: '{bottype[0]}'"));
                    return BotTypes.Invalid;
                }

                BotTypes BotType = BotTypes.Invalid;
                switch (bottype[1].Trim())
                {
                    case "Fighter":
                        BotType = BotTypes.Fighter;
                        break;
                    case "Freighter":
                        BotType = BotTypes.Freighter;
                        break;
                    case "Carrier":
                        BotType = BotTypes.Carrier;
                        break;
                    case "Station":
                        BotType = BotTypes.Station;
                        break;
                }

                return BotType;
            }
            catch (Exception Scrap)
            {
                RC.CubeGrid.LogError("[STATIC]BotBase.ReadBotType", Scrap);
                return BotTypes.Invalid;
            }
        }

        virtual protected void DebugWrite(string Source, string Message, string DebugPrefix = "BotBase.")
        {
            Grid.DebugWrite(DebugPrefix + Source, Message);
        }

        //Уникальное имя корабля, используется в основном для дебага
        public string GridName = "Unknown";
        virtual public bool Init(IMyRemoteControl RC = null)
        {
            this.RC = RC ?? Term.GetBlocksOfType<IMyRemoteControl>(collect: x => x.IsFunctional).FirstOrDefault();
            if (RC == null) return false;
            DroneName = DroneNameProvider;

            Antennae = Term.GetBlocksOfType<IMyRadioAntenna>(collect: x => x.IsFunctional);

            bool HasSetup = ParseSetup();
            if (!HasSetup) return false;

            AISessionCore.AddDamageHandler(Grid, (Block, Damage) => OnDamaged(Block, Damage));
            Grid.OnBlockAdded += (Block) => OnBlockPlaced(Block);
            //Grid.OnBlockIntegrityChanged
            //Grid.OnBlockRemoved +=
            OwnerFaction = Grid.GetOwnerFaction(RecalculateOwners: true);
            BotOperable = true;

            //Определяем имя корабля\станции
            GridName = String.Format("({0}){1}_{2}", OwnerFaction.Tag, Grid.DisplayName, Grid.EntityId.ToString());

            return true;
        }

        virtual public void RecompilePBs()
        {
            foreach (IMyProgrammableBlock PB in Term.GetBlocksOfType<IMyProgrammableBlock>())
            {
                PB.Recompile();
            }
        }

        virtual protected void ReactOnDamage(IMySlimBlock Block, MyDamageInformation Damage, TimeSpan TruceDelay, out IMyPlayer Damager)
        {
            Damager = null;
            try
            {
                if (Damage.IsMeteor())
                {
                    Grid.DebugWrite("ReactOnDamage", "Grid was damaged by meteor. Ignoring.");
                    return;
                }

                if (Damage.IsThruster())
                {
                    if (Block != null && !Block.IsDestroyed)
                    {
                        Grid.DebugWrite("ReactOnDamage", "Grid was slighly damaged by thruster. Ignoring.");
                        return;
                    }
                }

                try
                {
                    if (Damage.IsDoneByPlayer(out Damager) && Damager != null)
                    {
                        try
                        {
                            Grid.DebugWrite("ReactOnDamage", $"Grid is damaged by player {Damager.DisplayName}. Trying to activate alert.");
                            RegisterHostileAction(Damager, TruceDelay);
                        }
                        catch (Exception Scrap)
                        {
                            Grid.LogError("ReactOnDamage.GetDamagerFaction", Scrap);
                        }
                    }
                    else Grid.DebugWrite("ReactOnDamage", "Grid is damaged, but damage source is not recognized as player.");
                }
                catch (Exception Scrap)
                {
                    Grid.LogError("ReactOnDamage.IsDamageDoneByPlayer", Scrap);
                }
            }
            catch (Exception Scrap)
            {
                Grid.LogError("ReactOnDamage", Scrap);
            }
        }

        virtual protected void BlockPlacedHandler(IMySlimBlock Block)
        {
            if (Block == null) return;

            try
            {
                IMyPlayer Builder = null;
                IMyFaction Faction = null;
                if (Block.IsPlayerBlock(out Builder))
                {
                    Faction = Builder.GetFaction();
                    if (Faction != null)
                    {
                        RegisterHostileAction(Faction, CalmdownTime);
                    }
                }
            }
            catch (Exception Scrap)
            {
                Grid.LogError("BlokPlaedHandler", Scrap);
            }
        }

        virtual protected void RegisterHostileAction(IMyPlayer Player, TimeSpan TruceDelay)
        {
            try
            {
                #region Sanity checks
                if (Player == null)
                {
                    Grid.DebugWrite("RegisterHostileAction", "Error: Damager is null.");
                    return;
                };

                if (OwnerFaction == null)
                {
                    OwnerFaction = Grid.GetOwnerFaction();
                }

                if (OwnerFaction == null || !OwnerFaction.IsNPC())
                {
                    Grid.DebugWrite("RegisterHostileAction", $"Error: {(OwnerFaction == null ? "can't find own faction" : "own faction isn't recognized as NPC.")}");
                    return;
                }
                #endregion

                IMyFaction HostileFaction = Player.GetFaction();
                if (HostileFaction == null)
                {
                    Grid.DebugWrite("RegisterHostileAction", "Error: can't find damager's faction");
                    return;
                }
                else if (HostileFaction == OwnerFaction)
                {
                    OwnerFaction.Kick(Player);
                    return;
                }

                AISessionCore.DeclareWar(OwnerFaction, HostileFaction, TruceDelay);
                if (OwnerFaction.IsLawful())
                {
                    AISessionCore.DeclareWar(Diplomacy.Police, HostileFaction, TruceDelay);
                    AISessionCore.DeclareWar(Diplomacy.Army, HostileFaction, TruceDelay);
                }
            }
            catch (Exception Scrap)
            {
                LogError("RegisterHostileAction", Scrap);
            }
        }

        virtual protected void RegisterHostileAction(IMyFaction HostileFaction, TimeSpan TruceDelay)
        {
            try
            {
                if (HostileFaction == null)
                {
                    Grid.DebugWrite("RegisterHostileAction", "Error: can't find damager's faction");
                    return;
                }
                AISessionCore.DeclareWar(OwnerFaction, HostileFaction, TruceDelay);
                if (OwnerFaction.IsLawful())
                {
                    AISessionCore.DeclareWar(Diplomacy.Police, HostileFaction, TruceDelay);
                    AISessionCore.DeclareWar(Diplomacy.Army, HostileFaction, TruceDelay);
                }
            }
            catch (Exception Scrap)
            {
                LogError("RegisterHostileAction", Scrap);
            }
        }

        //Данные о состоянии сетки...
        public struct GridStatusData {
            public float Weapon_Left;
            public float EnergyBlock_Left;
            public float Gyro_Left;
            public float Total_Health_Left;
            public float RC_HP_left;
            public float Total_Blocks;
            public float Total_Health;
        }

        //Проверяем корабль на критические повреждения после которых нужно срочно отступить, используется как правило для боевых дронов
        public GridStatusData GetStatusData() {
            
            float Weapon_Left = 0f;
            float EnergyBlock_Left = 0f;
            float Gyro_Left = 0f;
            float Total_Health = 0f;
            int Total_Blocks = 0;
            float Total_Health_Left = 0f;

            //Получаем из сетки все блоки
            List<IMySlimBlock> tmp_block_list = new List<IMySlimBlock>();
            Grid.GetBlocks(tmp_block_list);

            for (int i = 0; i < tmp_block_list.Count; i++) {

                //Получаем "базовый" блок...
                IMySlimBlock tmp_block = tmp_block_list[i];

                //Получаем "здоровье" базового блока...
                float tmp_health = tmp_block.GetHealth();
                //Собираем информацию о всебх блоках чтобы потом высчитать общий процент "здоровья"
                Total_Blocks += 1;
                Total_Health += tmp_health;

                #region Определение типа блока
                //Распределяем приоритеты блоков в зависимости от их опасности\важности
                if (tmp_block.FatBlock is IMyLargeMissileTurret) {
                    IMyLargeMissileTurret tmp_Target = tmp_block.FatBlock as IMyLargeMissileTurret;
                    if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                        Weapon_Left += tmp_health;  
                    }
                }
                else if (tmp_block.FatBlock is IMySmallMissileLauncher) {
                    IMySmallMissileLauncher tmp_Target = tmp_block.FatBlock as IMySmallMissileLauncher;
                    if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                        Weapon_Left += tmp_health; 
                    }
                }
                else if (tmp_block.FatBlock is IMySmallMissileLauncherReload) {
                    IMySmallMissileLauncherReload tmp_Target = tmp_block.FatBlock as IMySmallMissileLauncherReload;
                    if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                        Weapon_Left += tmp_health; 
                    }
                }
                else if (tmp_block.FatBlock is IMyLargeGatlingTurret) {
                    IMyLargeGatlingTurret tmp_Target = tmp_block.FatBlock as IMyLargeGatlingTurret;
                    if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                        Weapon_Left += tmp_health; 
                    }
                }
                else if (tmp_block.FatBlock is IMySmallGatlingGun) {
                    IMySmallGatlingGun tmp_Target = tmp_block.FatBlock as IMySmallGatlingGun;
                    if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                        Weapon_Left += tmp_health; 
                    }
                }
                else if (tmp_block.FatBlock is IMyShipGrinder) {
                    IMyShipGrinder tmp_Target = tmp_block.FatBlock as IMyShipGrinder;
                    if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                        Weapon_Left += tmp_health; 
                    }
                }
                else if (tmp_block.FatBlock is IMyReactor) {
                    IMyReactor tmp_Target = tmp_block.FatBlock as IMyReactor;
                    if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                        EnergyBlock_Left += tmp_health; 
                    }
                }
                else if (tmp_block.FatBlock is IMyBatteryBlock) {
                    IMyBatteryBlock tmp_Target = tmp_block.FatBlock as IMyBatteryBlock;
                    if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                        EnergyBlock_Left += tmp_health; 
                    }
                }
                else if (tmp_block.FatBlock is IMyGyro) {
                    IMyGyro tmp_Target = tmp_block.FatBlock as IMyGyro;
                    if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                        Gyro_Left += tmp_health; 
                    }
                }
                #endregion

            }

            
            GridStatusData result_data = new GridStatusData();
            //Считаем итоговый процент здоровья корабля, тупо средний по всем блокам
            result_data.Total_Health_Left = Total_Health / Total_Blocks;
            //Получаем здоровье контрольного блока
            result_data.RC_HP_left = RC.SlimBlock.GetHealth();
            //Заполняем оставшиеся данные
            result_data.Weapon_Left = Weapon_Left;
            result_data.EnergyBlock_Left = EnergyBlock_Left;
            result_data.Gyro_Left = Gyro_Left;
            result_data.Total_Blocks = Total_Blocks;
            result_data.Total_Health = Total_Health;

            return result_data;
        }

        public struct CustomTargetData {
            public Ingame.MyDetectedEntityInfo _EntityInfo;
            public MyEntity _Entity;
            public int _Priority;
            public double _Distance;
            public string _Name;
        }

        //Ищем враждебные объекты вокруг корабля...
        protected List<CustomTargetData> LookAround(float Radius, Func<Ingame.MyDetectedEntityInfo, bool> Filter = null) {
            List<CustomTargetData> RadarData = new List<CustomTargetData>();
            BoundingSphereD LookaroundSphere = new BoundingSphereD(GridPosition, Radius);

            List<IMyEntity> EntitiesAround = MyAPIGateway.Entities.GetTopMostEntitiesInSphere(ref LookaroundSphere);
            EntitiesAround.RemoveAll(x => x == Grid || GridPosition.DistanceTo(x.GetPosition()) < GridRadius * 1.5);

            long OwnerID;
            if (OwnerFaction != null) {
                OwnerID = OwnerFaction.FounderId;
                Grid.DebugWrite("LookAround", "Found owner via faction owner");
            }
            else {
                OwnerID = RC.OwnerId;
                Grid.DebugWrite("LookAround", "OWNER FACTION NOT FOUND, found owner via RC owner");
            }

            foreach (IMyEntity DetectedEntity in EntitiesAround) {

                Ingame.MyDetectedEntityInfo RadarDetectedEntity = MyDetectedEntityInfoHelper.Create(DetectedEntity as MyEntity, OwnerID);
                if (Filter == null ? true : (Filter(RadarDetectedEntity))) {

                    IMyCubeGrid tmp_grid = DetectedEntity as IMyCubeGrid;
                    
                    //Если полученный объект является сеткой то это станция\корабль и нужно его просканировать чтобы получить самые актуальные цели
                    if (tmp_grid != null) { 

                        //int tg_block_count = (tmp_grid as MyCubeGrid).BlocksCount;

                        //Получаем из сетки все блоки с логикой, т.е. турели, контейнеры, кресла пилота и так далее и перебираем их...
                        List<IMySlimBlock> tmp_block_list = new List<IMySlimBlock>();
                        tmp_grid.GetBlocks(tmp_block_list, x => x.FatBlock != null);
                        for (int i = 0; i < tmp_block_list.Count; i++) {

                            //Получаем блок...
                            IMySlimBlock tmp_block = tmp_block_list[i];

                            
                            string tmp_b_name = "Unnamed";
                            int prioroty = 0;

                            #region Распределение приоритетов для блоков
                            //Распределяем приоритеты блоков в зависимости от их опасности\важности
                            if (tmp_block.FatBlock is IMyLargeMissileTurret) {
                                IMyLargeMissileTurret tmp_Target = tmp_block.FatBlock as IMyLargeMissileTurret;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 1;
                                }
                            }
                            else if (tmp_block.FatBlock is IMySmallMissileLauncher) {
                                IMySmallMissileLauncher tmp_Target = tmp_block.FatBlock as IMySmallMissileLauncher;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 2;
                                }
                            }
                            else if (tmp_block.FatBlock is IMySmallMissileLauncherReload) {
                                IMySmallMissileLauncherReload tmp_Target = tmp_block.FatBlock as IMySmallMissileLauncherReload;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 2;
                                }
                            }
                            else if (tmp_block.FatBlock is IMyLargeGatlingTurret) {
                                IMyLargeGatlingTurret tmp_Target = tmp_block.FatBlock as IMyLargeGatlingTurret;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 2;
                                }
                            }
                            else if (tmp_block.FatBlock is IMySmallGatlingGun) {
                                IMySmallGatlingGun tmp_Target = tmp_block.FatBlock as IMySmallGatlingGun;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 3;
                                }
                            }
                            else if (tmp_block.FatBlock is IMyLargeInteriorTurret) {
                                IMyLargeInteriorTurret tmp_Target = tmp_block.FatBlock as IMyLargeInteriorTurret;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 3;
                                }
                            }
                            else if (tmp_block.FatBlock is IMyShipGrinder) {
                                IMyShipGrinder tmp_Target = tmp_block.FatBlock as IMyShipGrinder;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 3;
                                }
                            }
                            else if (tmp_block.FatBlock is IMyRemoteControl) {
                                IMyRemoteControl tmp_Target = tmp_block.FatBlock as IMyRemoteControl;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 3;
                                }
                            }
                            else if (tmp_block.FatBlock is IMyCockpit) {
                                IMyCockpit tmp_Target = tmp_block.FatBlock as IMyCockpit;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 3;
                                }
                            }
                            else if (tmp_block.FatBlock is IMyReactor) {
                                IMyReactor tmp_Target = tmp_block.FatBlock as IMyReactor;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 4;
                                }
                            }
                            else if (tmp_block.FatBlock is IMyBatteryBlock) {
                                IMyBatteryBlock tmp_Target = tmp_block.FatBlock as IMyBatteryBlock;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 4;
                                }
                            }
                            else if (tmp_block.FatBlock is IMySolarPanel) {
                                IMySolarPanel tmp_Target = tmp_block.FatBlock as IMySolarPanel;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 4;
                                }
                            }
                            else if (tmp_block.FatBlock is IMyMedicalRoom) {
                                IMyMedicalRoom tmp_Target = tmp_block.FatBlock as IMyMedicalRoom;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 4;
                                }
                            }
                            else if (tmp_block.FatBlock is IMyCargoContainer) {
                                IMyCargoContainer tmp_Target = tmp_block.FatBlock as IMyCargoContainer;
                                if (tmp_Target.IsFunctional && tmp_Target.IsWorking) {
                                    tmp_b_name = tmp_Target.CustomName + "_" + tmp_Target.EntityId;
                                    prioroty = 5;
                                }
                            }
                            #endregion

                            //Если блок представляет какой-либо интерес его приоритет будет выше нуля...
                            if (prioroty > 0) {

                                //Заполняем данные о возможной цели...
                                Ingame.MyDetectedEntityInfo TargetEntity = MyDetectedEntityInfoHelper.Create(tmp_block.FatBlock as MyEntity, OwnerID);
                                CustomTargetData tmp_trg_data = new CustomTargetData();
                                tmp_trg_data._Name = tmp_b_name;
                                tmp_trg_data._Entity = tmp_block.FatBlock as MyEntity;
                                tmp_trg_data._EntityInfo = TargetEntity;
                                tmp_trg_data._Distance = GridPosition.DistanceTo(TargetEntity.Position);
                                tmp_trg_data._Priority = prioroty;

                                //Если дистанция до блока больше 1000 метров(максимальная дальность наведения турелей) то прибавляем к его приоритету 100 единиц...
                                //...отправляя его тем самым на дно списка и дрон будет лететь не к блоку а к игроку а затем уже будет агриться на блоки рядом с игроком
                                if (tmp_trg_data._Distance > 1000f) { tmp_trg_data._Priority += 100; }

                                RadarData.Add(tmp_trg_data); 
                                    
                                if (AISessionCore.AI_Debug) {
                                    Digi.Log.Info(String.Format("[TargetGrid: {0}_{1} - TargetBlockType: {2}, Name: {3}, Prioroty: {4}, Distance: {5}]",
                                        tmp_grid.DisplayName,
                                        tmp_grid.EntityId,
                                        tmp_block.FatBlock.GetType().ToString(),
                                        tmp_trg_data._Name,
                                        tmp_trg_data._Priority,
                                        tmp_trg_data._Distance
                                    ));
                                }
                            }
                        }

                    }
                    //...во всех остальных случаях это как правило игрок...
                    else if (tmp_grid == null) { 

                        CustomTargetData tmp_trg = new CustomTargetData();
                        tmp_trg._Name = DetectedEntity.DisplayName + "_" + DetectedEntity.EntityId;
                        tmp_trg._EntityInfo = RadarDetectedEntity;
                        tmp_trg._Entity = DetectedEntity as MyEntity;
                        tmp_trg._Distance = GridPosition.DistanceTo(RadarDetectedEntity.Position);
                        
                        //Если игрок слишком близко то задаем ему повышенный приоритет, в противном случае он нас почти не интересует
                        if (DetectedEntity is IMyCharacter && tmp_trg._Distance < 50f) { tmp_trg._Priority = 1; }
                        else if (DetectedEntity is IMyCharacter && tmp_trg._Distance < 100f) { tmp_trg._Priority = 2; }
                        else { tmp_trg._Priority = 5; }

                        RadarData.Add(tmp_trg); 
                        if (AISessionCore.AI_Debug) {
                            Digi.Log.Info(String.Format("[TargetOther: {0} - TargetBlockType: {1}, Prioroty: {2}, Distance: {3}]",
                                tmp_trg._Name,
                                DetectedEntity.GetType().ToString(),
                                tmp_trg._Priority,
                                tmp_trg._Distance
                            ));
                        }
                    }
                    
                }
            }

            return RadarData;
        }

        protected List<CustomTargetData> LookForEnemies(float Radius, bool ConsiderNeutralsAsHostiles = false) {
            if (!ConsiderNeutralsAsHostiles) { return LookAround(Radius, x => x.Relationship == MyRelationsBetweenPlayerAndBlock.Enemies); }
            else { return LookAround(Radius, x => x.Relationship == MyRelationsBetweenPlayerAndBlock.Enemies || x.Relationship == MyRelationsBetweenPlayerAndBlock.Neutral); }
        }

        virtual protected List<IMyTerminalBlock> GetHackedBlocks() {
            List<IMyTerminalBlock> TerminalBlocks = new List<IMyTerminalBlock>();
            List<IMyTerminalBlock> HackedBlocks = new List<IMyTerminalBlock>();

            Term.GetBlocks(TerminalBlocks);

            foreach (IMyTerminalBlock Block in TerminalBlocks)
                if (Block.IsBeingHacked) HackedBlocks.Add(Block);

            return HackedBlocks;
        }

        virtual protected List<IMySlimBlock> GetDamagedBlocks() {
            List<IMySlimBlock> Blocks = new List<IMySlimBlock>();
            Grid.GetBlocks(Blocks, x => x.CurrentDamage > 10);
            return Blocks;
        }

        protected bool HasModdedThrusters => SpeedmoddedThrusters.Count > 0;
        protected List<IMyThrust> SpeedmoddedThrusters = new List<IMyThrust>();

        protected void ApplyThrustMultiplier(float ThrustMultiplier)
        {
            DemultiplyThrusters();
            foreach (IMyThrust Thruster in Term.GetBlocksOfType<IMyThrust>(collect: x => x.IsOwnedByNPC(AllowNobody: false, CheckBuilder: true)))
            {
                Thruster.ThrustMultiplier = ThrustMultiplier;
                Thruster.OwnershipChanged += Thruster_OnOwnerChanged;
                SpeedmoddedThrusters.Add(Thruster);
            }
        }

        protected void DemultiplyThrusters()
        {
            if (!HasModdedThrusters) return;
            foreach (IMyThrust Thruster in SpeedmoddedThrusters)
            {
                if (Thruster.ThrustMultiplier != 1) Thruster.ThrustMultiplier = 1;
            }
            SpeedmoddedThrusters.Clear();
        }

        private void Thruster_OnOwnerChanged(IMyTerminalBlock thruster)
        {
            try
            {
                IMyThrust Thruster = thruster as IMyThrust;
                if (Thruster == null) return;
                if (!Thruster.IsOwnedByNPC() && Thruster.ThrustMultiplier != 1) Thruster.ThrustMultiplier = 1;
            }
            catch (Exception Scrap)
            {
                Grid.DebugWrite("Thruster_OnOwnerChanged", $"{thruster.CustomName} OnOwnerChanged failed: {Scrap.Message}");
            }
        }

        abstract protected bool ParseSetup();

        abstract public void Main();

        virtual public void Shutdown()
        {
            Closed = true;
            if (HasModdedThrusters) DemultiplyThrusters();
            AISessionCore.RemoveDamageHandler(Grid);
        }

        public void LogError(string Source, Exception Scrap, string DebugPrefix = "BotBase.")
        {
            Grid.LogError(DebugPrefix + Source, Scrap);
        }
    }

    /*public sealed class InvalidBot : BotBase
    {
        static public readonly BotTypes BotType = BotTypes.None;

        public override bool Operable
        {
            get
            {
                return false;
            }
        }

        public InvalidBot(IMyCubeGrid Grid = null) : base(Grid)
        {
        }

        public override bool Init(IMyRemoteControl RC = null)
        {
            return false;
        }

        public override void Main()
        {
            // Empty
        }

        protected override bool ParseSetup()
        {
            return false;
        }
    }*/
}